import { Subject, BehaviorSubject } from 'rxjs';

import { AxiosInstance } from 'axios';

import * as Electron from 'electron';

import { StatusCode } from '@ucap/domain-status';

import {
  NativeService,
  UpdateCheckConfig,
  UpdateInfo,
  NativeType,
  WindowState,
  WindowIdle,
  NotificationRequest,
  AppInitInfo
} from '@ucap/native';
import {
  PlatformChannel,
  FileChannel,
  WindowChannel,
  IdleChannel,
  AppChannel,
  ChatChannel,
  MessageChannel
} from '../types/channel.type';

const listenersOfChannel = [
  WindowChannel.onState$,
  WindowChannel.onFocus$,
  IdleChannel.onState$,
  AppChannel.onUpdate$,
  AppChannel.onLogout$,
  AppChannel.onStatus$,
  AppChannel.onShowSetting$,
  AppChannel.onExit$,
  ChatChannel.onOpen$,
  MessageChannel.onOpen$
];

export class ElectronNativeService extends NativeService {
  private ipcRenderer: typeof Electron.ipcRenderer;
  private webFrame: typeof Electron.webFrame;
  private remote: typeof Electron.remote;
  private shell: typeof Electron.shell;

  private platform: string;

  constructor(private axios: AxiosInstance) {
    super();

    this.ipcRenderer = (window as any).require('electron').ipcRenderer;
    this.remote = (window as any).require('electron').remote;
    this.shell = (window as any).require('electron').shell;
    this.webFrame = (window as any).require('electron').webFrame;

    this.platform = (window as any).require('os').platform();
  }

  platform_nativeType(): Promise<NativeType> {
    return new Promise<any>(async (resolve, reject) => {
      try {
        resolve(NativeType.Electron);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_networkInfo(): Promise<any> {
    return new Promise<any>(async (resolve, reject) => {
      try {
        const networkInfo = await this.ipcRenderer.invoke(
          PlatformChannel.networkInfo
        );
        resolve(networkInfo);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_execute(executableName: string): Promise<number> {
    return new Promise<number>(async (resolve, reject) => {
      try {
        const id = await this.ipcRenderer.invoke(
          PlatformChannel.execute,
          executableName
        );
        resolve(id);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_openDefaultBrowser(
    url: string,
    options?: {
      name?: string;
      features?: string;
      replace?: boolean;
    }
  ): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(
          PlatformChannel.openDefaultBrowser,
          url,
          options
        );

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_readFromClipboard(): Promise<{
    text?: string;
    rtf?: string;
    html?: string;
    image?: Buffer;
    imageDataUrl?: string;
  }> {
    return new Promise<{
      text?: string;
      rtf?: string;
      html?: string;
      image?: Buffer;
      imageDataUrl?: string;
    }>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          PlatformChannel.readFromClipboard
        );

        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }

  file_save(
    buffer: Buffer,
    fileName: string,
    mimeType: string,
    path?: string
  ): Promise<string> {
    return new Promise<string>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          FileChannel.save,
          buffer,
          fileName,
          mimeType,
          path
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_read(path: string): Promise<Buffer> {
    return new Promise<Buffer>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(FileChannel.read, path);
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_openFolder(folderPath?: string, make?: boolean): Promise<boolean> {
    return new Promise<boolean>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          FileChannel.openFolder,
          folderPath,
          make
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_openItem(filePath?: string): Promise<boolean> {
    return new Promise<boolean>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          FileChannel.openItem,
          filePath
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_path(
    name: 'home' | 'documents' | 'downloads' | 'music' | 'pictures' | 'videos',
    ...appendPaths: string[]
  ): Promise<string> {
    return new Promise<string>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          FileChannel.path,
          name,
          appendPaths
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_selectForOpen(option: {
    title?: string;
    defaultPath?: string;
    filters?: {
      extensions: string[];
      name: string;
    }[];
    properties?: Array<
      | 'openFile'
      | 'openDirectory'
      | 'multiSelections'
      | 'showHiddenFiles'
      | 'createDirectory'
      | 'promptToCreate'
      | 'noResolveAliases'
      | 'treatPackageAsDirectory'
    >;
    message?: string;
  }): Promise<string> {
    return new Promise<string>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          FileChannel.selectForOpen,
          option
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_selectForSave(option: {
    title?: string;
    defaultPath?: string;
    filters?: {
      extensions: string[];
      name: string;
    }[];
    message?: string;
  }): Promise<{
    canceled: boolean;
    filePath: string;
  }> {
    return new Promise<{
      canceled: boolean;
      filePath: string;
    }>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          FileChannel.selectForSave,
          option
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }

  window_state(): Promise<WindowState> {
    return new Promise<WindowState>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(WindowChannel.state);
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  window_onState$(): BehaviorSubject<WindowState> {
    if (!this._window_onStateSubject) {
      this.ipcRenderer.on(
        WindowChannel.onState$,
        (event: any, state: WindowState) => {
          if (!!this._window_onStateSubject) {
            this._window_onStateSubject.next(state);
          }
        }
      );

      this.window_state().then((state) => {
        this._window_onStateSubject.next(state);
      });
    }

    return super.window_onState$();
  }
  window_focused(): Promise<boolean> {
    return new Promise<boolean>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(WindowChannel.focused);
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  window_onFocus$(): BehaviorSubject<boolean> {
    if (!this._window_onFocusSubject) {
      this.ipcRenderer.on(
        WindowChannel.onFocus$,
        (event: any, focused: boolean) => {
          if (!!this._window_onFocusSubject) {
            this._window_onFocusSubject.next(focused);
          }
        }
      );

      this.window_focused().then((focused) => {
        this._window_onFocusSubject.next(focused);
      });
    }

    return super.window_onFocus$();
  }
  window_close(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const currentWindow = this.remote.getCurrentWindow();
        if (!!currentWindow) {
          currentWindow.hide();
        }
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  window_minimize(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const currentWindow = this.remote.getCurrentWindow();
        if (!!currentWindow) {
          currentWindow.minimize();
        }
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  window_maximize(altKey: boolean = false): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const currentWindow = this.remote.getCurrentWindow();
        if (!!currentWindow) {
          if ('darwin' === this.platform) {
            if (altKey) {
              if (currentWindow.isMaximizable()) {
                if (currentWindow.isMaximized()) {
                  currentWindow.unmaximize();
                } else {
                  currentWindow.maximize();
                }
              }
            } else {
              currentWindow.setFullScreen(!currentWindow.isFullScreen());
            }
          } else {
            if (currentWindow.isMaximized()) {
              currentWindow.unmaximize();
            } else {
              currentWindow.maximize();
            }
          }
        }

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  idle_startCheck(limitTime: number): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const started = await this.ipcRenderer.invoke(IdleChannel.started);
        if (started) {
          resolve();
          return;
        }

        await this.ipcRenderer.invoke(IdleChannel.startCheck, limitTime);

        this.ipcRenderer.on(
          IdleChannel.onState$,
          this._idle_onState$Callback.bind(this)
        );

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  _idle_onState$Callback(event: any, idleState: WindowIdle) {
    if (!!this._idle_onStateSubject) {
      this._idle_onStateSubject.next(idleState);
    }
  }
  idle_onState$(): BehaviorSubject<WindowIdle> {
    return super.idle_onState$();
  }
  idle_stopCheck(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const started = await this.ipcRenderer.invoke(IdleChannel.started);
        if (!started) {
          resolve();
          return;
        }

        await this.ipcRenderer.invoke(IdleChannel.stopCheck);

        this.ipcRenderer.off(
          IdleChannel.onState$,
          this._idle_onState$Callback.bind(this)
        );

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  idle_changeLimitTime(limitTime: number): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(IdleChannel.changeLimitTime, limitTime);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_version(): Promise<string> {
    return new Promise<any>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(AppChannel.version);
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postInit(): Promise<AppInitInfo> {
    return new Promise<AppInitInfo>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(AppChannel.postInit);
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postLogin(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.postLogin);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postLogout(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.postLogout);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postDestroy(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.postDestroy);

        for (const channel of listenersOfChannel) {
          this.ipcRenderer.removeAllListeners(channel);
        }

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_changeAutoLaunch(autoLaunch: boolean): Promise<boolean> {
    return new Promise<boolean>(async (resolve, reject) => {
      try {
        const data = await this.ipcRenderer.invoke(
          AppChannel.changeAutoLaunch,
          autoLaunch
        );
        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  }
  app_showNotify(req: NotificationRequest): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.showNotify, req);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_closeAllNotify(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.closeAllNotify);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_checkForUpdates(currentVersion: string): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(
          AppChannel.checkForUpdates,
          currentVersion
        );
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_applyInstantUpdates(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.applyInstantUpdates);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_startCheckForUpdate(config: UpdateCheckConfig): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const started = await this.ipcRenderer.invoke(
          AppChannel.startedCheckForUpdate
        );
        if (started) {
          resolve();
          return;
        }

        await this.ipcRenderer.invoke(AppChannel.startCheckForUpdate, config);

        this.ipcRenderer.on(
          AppChannel.onUpdate$,
          this._app_onUpdate$Callback.bind(this)
        );

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  _app_onUpdate$Callback(event: any, info: UpdateInfo) {
    if (!!this._app_onUpdateSubject) {
      this._app_onUpdateSubject.next(info);
    }
  }
  app_onUpdate$(): BehaviorSubject<UpdateInfo> {
    return super.app_onUpdate$();
  }
  app_stopCheckForUpdate(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        const started = await this.ipcRenderer.invoke(
          AppChannel.startedCheckForUpdate
        );
        if (!started) {
          resolve();
          return;
        }

        await this.ipcRenderer.invoke(AppChannel.stopCheckForUpdate);

        this.ipcRenderer.off(
          AppChannel.onUpdate$,
          this._app_onUpdate$Callback.bind(this)
        );

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_exit(): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(AppChannel.exit);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_onLogout$(): Subject<void> {
    if (!this._app_onLogoutSubject) {
      this.ipcRenderer.on(AppChannel.onLogout$, (event: any) => {
        if (!!this._app_onLogoutSubject) {
          this._app_onLogoutSubject.next();
        }
      });
    }

    return super.app_onLogout$();
  }
  app_onStatus$(): BehaviorSubject<StatusCode> {
    if (!this._app_onStatusSubject) {
      this.ipcRenderer.on(
        AppChannel.onStatus$,
        (event: any, code: StatusCode) => {
          if (!!this._app_onStatusSubject) {
            this._app_onStatusSubject.next(code);
          }
        }
      );
    }

    return super.app_onStatus$();
  }
  app_onShowSetting$(): Subject<void> {
    if (!this._app_onShowSettingSubject) {
      this.ipcRenderer.on(AppChannel.onShowSetting$, (event: any) => {
        if (!!this._app_onShowSettingSubject) {
          this._app_onShowSettingSubject.next();
        }
      });
    }

    return super.app_onShowSetting$();
  }

  app_onExit$(): Subject<void> {
    if (!this._app_onExitSubject) {
      this.ipcRenderer.on(AppChannel.onExit$, (event: any) => {
        if (!!this._app_onExitSubject) {
          this._app_onExitSubject.next();
        }
      });
    }

    return super.app_onExit$();
  }

  chat_openRoom(openUrl: string): Promise<void> {
    return new Promise<void>(async (resolve, reject) => {
      try {
        await this.ipcRenderer.invoke(ChatChannel.openRoom, openUrl);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  chat_onOpen$(): Subject<string> {
    if (!this._chat_onOpenSubject) {
      this.ipcRenderer.on(ChatChannel.onOpen$, (event: any, seq: string) => {
        if (!!this._chat_onOpenSubject) {
          this._chat_onOpenSubject.next(seq);
        }
      });
    }

    return super.chat_onOpen$();
  }

  message_onOpen$(): Subject<string> {
    if (!this._message_onOpenSubject) {
      this.ipcRenderer.on(MessageChannel.onOpen$, (event: any, seq: string) => {
        if (!!this._message_onOpenSubject) {
          this._message_onOpenSubject.next(seq);
        }
      });
    }

    return super.message_onOpen$();
  }
}
